/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

package org.openisis;

import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;
import java.util.Iterator;
import java.util.NoSuchElementException;


/**
	This class is the native implementation of class Db.
<h3>Installation</h3>
	The shared library "openjsis" must be loadable.
	On a UNIX System, make sure that the file libopenjsis.so
	resides in a directory listed by the environment variable
	LD_LIBRARY_PATH or in /etc/ld.so.conf. 
	If you just drop it in /usr/lib or similar dir,
	you may have to run a command like ldconfig.
<p>
	On Windows, put openjsis.dll somewhere in the PATH.
<p>
	$Id: NativeDb.java,v 1.2 2002/10/28 11:46:42 kripke Exp $
	@version $Revision: 1.2 $
	@author  $Author: kripke $
*/
public class NativeDb extends Db {

	private static synchronized native
	short nopen ( String dbname, String[] args );

	/** read a record by rowid.
	 *	The native implementation gives character data as byte[],
	 *	since we need to employ java encoding mechanisms anyway.
	 *	The tag is packed as four bytes little endian.
	 *	@param db a db number as returned by open
	 *	@param rowid the number (aka mfn) of the row
	 *	@param tag in which to look for txt, negative for any
	 *	@param txt to grep from record, null for no grep
	 *	@return an array of Objects:
	 *		the first is an array of ints filled with rowid an tags
	 *		the others are byte[] with field character data in encoding
	 *	@see #encoding
	*/
	private static synchronized native
	Object[] nreadRow ( short db, int rowid, int tag, byte[] txt );

	private static synchronized native
	Object[] nTerms ( short db, byte[] term, byte[] start );

	private static synchronized native
	int[] nsearch ( short db, byte[] key, int mode );

	static {
		// TODO: try some explicit paths like our jar's location with System.load
		// in order to not have to export LD_LIBRARY_PATH or so
		System.loadLibrary( "openjsis" );
	}

	public byte[] cvt ( String txt ) {
		try {
			return txt.getBytes(encoding);
		} catch ( UnsupportedEncodingException e ) {
			// shouldn't happen -- our fault
			throw new Error( "failed to check for support of encoding "+encoding );
		}
	}

	public String cvt ( byte[] b ) {
		try {
			// System.err.println( "got "+b.length+" bytes "+b );
			for ( int l=b.length; 0<l; l-- ) {
				String s = new String( b, 0, l, encoding );
				if ( null != s && 0 != s.length() )
					return s;
				for ( int i=0; i<l; i++ )
					System.err.print(
						(
						b[i] > 0 ? "0" : // 0... self-contained
						b[i] < -64 ? "f" : // 10... six follower bits
						b[i] < -32 ? "1" : // 110... five bits, one follower
						b[i] < -16 ? "2" : // 1110... four bits, two followers
						"b" // bad
						) + "("+b[i]+")"
					);
				return "";
			}
			return "";
		} catch ( UnsupportedEncodingException e ) {
			// shouldn't happen -- our fault
			throw new Error( "failed to check for support of encoding "+encoding );
		}
	}


	/** Number of this db.
	*/
	public final short db;
	

	public NativeDb ( String enc, String dbname, String[] args )
		throws IOException
	{
		super( enc );
		db = nopen( dbname, args );
		if ( 0 > db )
			throw new FileNotFoundException();
	}

	Rec mkRow ( Object[] buf ) {
		if ( null == buf )
			return null;
		int[] ids = (int[])buf[0];
		int rowid = ids[0];
		Field[] f = new Field[ buf.length-1 ];
		for ( int i=1; i<buf.length; i++ )
			f[i-1] = new Field( ids[i], cvt( (byte[])buf[i] ) );
		return new Rec( rowid, f );
	}	// readrow

	public Rec readRow ( int rowid ) {
		return mkRow( nreadRow( db, rowid, -1, null ) );
	}

	public Rec scanRow ( int rowid, int tag, String txt ) {
		return mkRow( nreadRow( db, rowid, tag, cvt(txt) ) );
	}

	public int[] search ( String key, int mode ) {
		return nsearch( db, cvt(key), mode );
	}

	public class TermIterator implements Iterator {
		byte[] _prefix;
		Object[] buf;
		int nxt;

		public TermIterator ( String prefix ) {
			_prefix = cvt(prefix+"$");
		}

		public boolean hasNext () {
			if ( null != buf && nxt < buf.length ) return true;
			if ( nxt < 0 ) return false;
			nxt = -1; // mark, just in case we bomb
			buf = nTerms( db, _prefix,
				null == buf ? null : (byte[])buf[buf.length-1] );
			if ( null == buf || 1 >= buf.length ) {
				buf = null;
				return false;
			}
			nxt = 1;
			return true;
		}

		public Object next () {
			if ( ! hasNext() ) throw new NoSuchElementException();
			return cvt( (byte[])buf[nxt++] );
		}

		public void remove () { throw new UnsupportedOperationException(); }

	}

	public Iterator terms ( String prefix ) {
		return new TermIterator( prefix );
	}

}	// NativeDb
